@@ -1,5 +1,5 @@
 /******************************************************************************
- * Copyright 1994-2005,2006 by Thomas E. Dickey                               *
+ * Copyright 1994-2008,2009 by Thomas E. Dickey                               *
  * All Rights Reserved.                                                       *
  *                                                                            *
  * Permission to use, copy, modify, and distribute this software and its      *
@@ -20,7 +20,7 @@
  ******************************************************************************/
 
 #ifndef	NO_IDENT
-static const char *Id = "$Id: diffstat.c,v 1.42 2006/07/02 15:58:37 tom Exp $";
+static const char *Id = "$Id: diffstat.c,v 1.47 2009/03/29 18:04:40 tom Exp $";
 #endif
 
 /*
@@ -28,6 +28,25 @@ static const char *Id = "$Id: diffstat.c,v 1.42 2006/07/02 15:58:37 tom Exp $";
  * Author:	T.E.Dickey
  * Created:	02 Feb 1992
  * Modified:
+ *		29 Mar 2009, modify to work with patch ".rej" files, which have
+ *			     no filename header (use the name of the ".rej"
+ *			     file if it is available).
+ *		29 Sep 2008, fix typo in usage message.
+ *		06 Aug 2008, add "-m", "-S" and "-D" options.
+ *		05 Aug 2008, add "-q" option to suppress 0-files-changed
+ *			     message (patch by Greg Norris).
+ *		04 Sep 2007, add "-b" option to suppress binary-files (patch
+ *			     by Greg Norris).
+ *		26 Aug 2007, add "-d" option to show debugging traces, rather
+ *			     than by defining DEBUG.  Add check after
+ *			     unified-diff chunk to avoid adding non-diff text
+ *			     (report by Adrian Bunk).  Quote pathname passed
+ *			     in command to gzip/uncompress.  Add a check for
+ *			     default-diff output without the "diff" command
+ *			     supplied to provide filename, mark as "unknown".
+ *		16 Jul 2006, fix to avoid modifying which is being used by
+ *			     tsearch() for ordering the binary tree (report by
+ *			     Adrian Bunk).
  *		02 Jul 2006, do not ignore pathnames in /tmp/, since some tools
  *			     create usable pathnames for both old/new files
  *			     there (Debian #376086).  Correct ifdef for
@@ -142,13 +161,13 @@ static const char *Id = "$Id: diffstat.c,v 1.42 2006/07/02 15:58:37 tom Exp $";
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #else
-extern int atoi();
+extern int atoi(const char *);
 #endif
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #else
-extern int isatty();
+extern int isatty(int);
 #endif
 
 #ifdef HAVE_MALLOC_H
@@ -175,6 +194,9 @@ extern char *optarg;
 extern int optind;
 #endif
 
+#include <sys/types.h>
+#include <sys/stat.h>
+
 #if !defined(EXIT_SUCCESS)
 #define EXIT_SUCCESS 0
 #define EXIT_FAILURE 1
@@ -206,8 +228,12 @@ extern int optind;
 
 #define UC(c)   ((unsigned char)(c))
 
-#ifdef DEBUG
-#define TRACE(p) printf p
+#ifndef OPT_TRACE
+#define OPT_TRACE 1
+#endif
+
+#if OPT_TRACE
+#define TRACE(p) if (trace_opt) printf p
 #else
 #define TRACE(p)		/*nothing */
 #endif
@@ -228,38 +254,57 @@ typedef enum comment {
     Normal, Only, Binary
 } Comment;
 
-#define MARKS 3			/* each of +, - and ! */
+#define MARKS 4			/* each of +, - and ! */
 
-#define InsOf(p) (p)->count[0]	/* "+" count inserted lines */
-#define DelOf(p) (p)->count[1]	/* "-" count deleted lines */
-#define ModOf(p) (p)->count[2]	/* "!" count modified lines */
+typedef enum {
+    cInsert = 0,
+    cDelete,
+    cModify,
+    cEquals
+} Change;
 
-#define TotalOf(p) (InsOf(p) + DelOf(p) + ModOf(p))
+#define InsOf(p) (p)->count[cInsert]	/* "+" count inserted lines */
+#define DelOf(p) (p)->count[cDelete]	/* "-" count deleted lines */
+#define ModOf(p) (p)->count[cModify]	/* "!" count modified lines */
+#define EqlOf(p) (p)->count[cEquals]	/* "=" count unmodified lines */
+
+#define TotalOf(p) (InsOf(p) + DelOf(p) + ModOf(p) + EqlOf(p))
+#define for_each_mark(n) for (n = 0; n < num_marks; ++n)
 
 typedef struct _data {
     struct _data *link;
     char *name;			/* the filename */
     int base;			/* beginning of name if -p option used */
     Comment cmt;
-    long count[3];
+    int pending;
+    long chunks;		/* total number of chunks */
+    long chunk[MARKS];		/* counts for the current chunk */
+    long count[MARKS];		/* counts for the file */
 } DATA;
 
-static const char marks[MARKS + 1] = "+-!";
+static const char marks[MARKS + 1] = "+-!=";
 
 static DATA *all_data;
 static char *comment_opt = "";
+static char *path_opt = 0;
 static int format_opt = FMT_NORMAL;
 static int max_width;		/* the specified width-limit */
 static int merge_names = 1;	/* true if we merge similar filenames */
+static int merge_opt = 0;	/* true if we merge ins/del as modified */
 static int name_wide;		/* the amount reserved for filenames */
 static int names_only;		/* true if we list filenames only */
+static int num_marks = 3;	/* 3 or 4, according to "-P" option */
 static int show_progress;	/* if not writing to tty, show progress */
+static int path_dest;		/* true if path_opt is destination (patched) */
 static int plot_width;		/* the amount left over for histogram */
 static int prefix_opt = -1;	/* if positive, controls stripping of PATHSEP */
 static int round_opt = 0;	/* if nonzero, round data for histogram */
 static int table_opt = 0;	/* if nonzero, write table rather than plot */
+static int trace_opt = 0;	/* if nonzero, write debugging information */
 static int sort_names = 1;	/* true if we sort filenames */
-static int verbose = 0;		/* -q/-v options */
+static int verbose = 0;		/* -v option */
+static int quiet = 0;		/* -q option */
+static int suppress_binary = 0;	/* -b option */
 static long plot_scale;		/* the effective scale (1:maximum) */
 
 #ifdef HAVE_TSEARCH
@@ -288,6 +333,14 @@ xmalloc(size_t s)
     return p;
 }
 
+static int
+is_dir(const char *name)
+{
+    struct stat sb;
+    return (stat(name, &sb) == 0 &&
+	    S_ISDIR(sb.st_mode));
+}
+
 static void
 blip(int c)
 {
@@ -311,19 +364,47 @@ compare_data(const void *a, const void *b)
     return strcmp(p->name + p->base, q->name + q->base);
 }
 
+static void
+init_data(DATA * data, char *name, int base)
+{
+    memset(data, 0, sizeof(*data));
+    data->name = name;
+    data->base = base;
+    data->cmt = Normal;
+}
+
 static DATA *
 new_data(char *name, int base)
 {
     DATA *r = (DATA *) xmalloc(sizeof(DATA));
 
-    memset(r, 0, sizeof(*r));
-    r->name = new_string(name);
-    r->base = base;
-    r->cmt = Normal;
+    init_data(r, new_string(name), base);
 
     return r;
 }
 
+#ifdef HAVE_TSEARCH
+static DATA *
+add_tsearch_data(char *name, int base)
+{
+    DATA find;
+    DATA *result;
+    void *pp;
+
+    init_data(&find, name, base);
+    if ((pp = tfind(&find, &sorted_data, compare_data)) != 0) {
+	result = *(DATA **) pp;
+	return result;
+    }
+    result = new_data(name, base);
+    (void) tsearch(result, &sorted_data, compare_data);
+    result->link = all_data;
+    all_data = result;
+
+    return result;
+}
+#endif
+
 static DATA *
 find_data(char *name)
 {
@@ -331,7 +412,7 @@ find_data(char *name)
     DATA find;
     int base = 0;
 
-    TRACE(("find_data(%s)\n", name));
+    TRACE(("** find_data(%s)\n", name));
 
     /* Compute the base offset if the prefix option is used */
     if (prefix_opt >= 0) {
@@ -343,17 +424,9 @@ find_data(char *name)
 		break;
 	    base = s - name;
 	}
-	TRACE(("base set to %d\n", base));
+	TRACE(("** base set to %d\n", base));
     }
 
-    /*
-     * Setup parameter for compare_data().
-     */
-    memset(&find, 0, sizeof(find));
-    find.name = name;
-    find.base = base;
-    find.cmt = Normal;
-
     /* Insert into sorted list (usually sorted).  If we are not sorting or
      * merging names, we fall off the end and link the new entry to the end of
      * the list.  If the prefix option is used, the prefix is ignored by the
@@ -364,18 +437,11 @@ find_data(char *name)
      */
 #ifdef HAVE_TSEARCH
     if (use_tsearch) {
-	void *pp;
-	if ((pp = tfind(&find, &sorted_data, compare_data)) != 0) {
-	    p = *(DATA **) pp;
-	    return p;
-	}
-	r = new_data(name, base);
-	(void) tsearch(r, &sorted_data, compare_data);
-	r->link = all_data;
-	all_data = r;
+	r = add_tsearch_data(name, base);
     } else
 #endif
     {
+	init_data(&find, name, base);
 	for (p = all_data, q = 0; p != 0; q = p, p = p->link) {
 	    int cmp = compare_data(p, &find);
 	    if (merge_names && (cmp == 0))
@@ -403,12 +469,13 @@ delink(DATA * data)
 {
     DATA *p, *q;
 
-    TRACE(("delink '%s'\n", data->name));
+    TRACE(("** delink '%s'\n", data->name));
 
 #ifdef HAVE_TSEARCH
-    if (use_tsearch)
+    if (use_tsearch) {
 	if (tdelete(data, &sorted_data, compare_data) == 0)
 	    return 0;
+    }
 #endif
     for (p = all_data, q = 0; p != 0; q = p, p = p->link) {
 	if (p == data) {
@@ -465,6 +532,50 @@ edit_range(const char *s)
 	|| (sscanf(s, "%d%c", &first, temp) == 1);
 }
 
+/*
+ * Decode a range for default diff.
+ */
+static int
+decode_default(char *s,
+	       int *first, int *first_size,
+	       int *second, int *second_size)
+{
+    int rc = 0;
+    char *next;
+
+    if (isdigit(UC(*s))) {
+	*first_size = 1;
+	*second_size = 1;
+
+	*first = strtol(s, &next, 10);
+	if (next != 0 && next != s) {
+	    if (*next == ',') {
+		s = ++next;
+		*first_size = strtol(s, &next, 10) + 1 - *first;
+	    }
+	}
+	if (next != 0 && next != s) {
+	    switch (*next++) {
+	    case 'a':
+	    case 'c':
+	    case 'd':
+		s = next;
+		*second = strtol(s, &next, 10);
+		if (next != 0 && next != s) {
+		    if (*next == ',') {
+			s = ++next;
+			*second_size = strtol(s, &next, 10) + 1 - *second;
+		    }
+		}
+		if (next != 0 && next != s && *next == '\0')
+		    rc = 1;
+		break;
+	    }
+	}
+    }
+    return rc;
+}
+
 /*
  * Decode a range for unified diff.  Oddly, the comments in diffutils code
  * claim that both numbers are line-numbers.  However, inspection of the output
@@ -473,16 +584,20 @@ edit_range(const char *s)
 static int
 decode_range(const char *s, int *first, int *second)
 {
+    int rc = 0;
     char check;
-    if (sscanf(s, "%d,%d%c", first, second, &check) == 2) {
-	TRACE(("decode_range #1 first=%d, second=%d\n", *first, *second));
-	return 1;
-    } else if (sscanf(s, "%d%c", first, &check) == 1) {
-	*second = *first;	/* diffutils 2.7 does this */
-	TRACE(("decode_range #2 first=%d, second=%d\n", *first, *second));
-	return 1;
+
+    if (isdigit(UC(*s))) {
+	if (sscanf(s, "%d,%d%c", first, second, &check) == 2) {
+	    TRACE(("** decode_range #1 first=%d, second=%d\n", *first, *second));
+	    rc = 1;
+	} else if (sscanf(s, "%d%c", first, &check) == 1) {
+	    *second = *first;	/* diffutils 2.7 does this */
+	    TRACE(("** decode_range #2 first=%d, second=%d\n", *first, *second));
+	    rc = 1;
+	}
     }
-    return 0;
+    return rc;
 }
 
 static int
@@ -522,12 +637,12 @@ is_leaf(const char *theLeaf, const char *path)
 static char *
 do_merging(DATA * data, char *path, int *freed)
 {
-    TRACE(("do_merging(%s,%s) diffs:%d\n", data->name, path, HadDiffs(data)));
+    TRACE(("** do_merging(%s,%s) diffs:%d\n", data->name, path, HadDiffs(data)));
 
     *freed = 0;
     if (!HadDiffs(data)) {	/* the data was the first of 2 markers */
 	if (is_leaf(data->name, path)) {
-	    TRACE(("is_leaf: %s vs %s\n", data->name, path));
+	    TRACE(("** is_leaf: %s vs %s\n", data->name, path));
 	    *freed = delink(data);
 	} else if (can_be_merged(data->name)
 		   && can_be_merged(path)) {
@@ -536,17 +651,40 @@ do_merging(DATA * data, char *path, int *freed)
 	    unsigned n;
 	    int matched = 0;
 	    int diff = 0;
+	    int local = 0;
 
-	    /* strip suffixes such as ".orig", ".bak" */
+	    /*
+	     * Strip suffixes such as ".orig", ".bak", "~", etc.  The current
+	     * data for the merge is in 'path'.  The previous data (in
+	     * 'data->name') may also be a temporary filename (which would not
+	     * be merged since it has no apparent relationship to the current).
+	     */
 	    if (len1 > len2) {
 		if (!strncmp(data->name, path, len2)) {
-		    TRACE(("trimming data '%s' to '%.*s'\n",
+		    TRACE(("** trimming data '%s' to '%.*s'\n",
 			   data->name, (int) len2, data->name));
-		    data->name[len1 = len2] = EOS;
+		    len1 = len2;
+#ifdef HAVE_TSEARCH
+		    /*
+		     * If we are using tsearch(), make a local copy of the data
+		     * so we can trim it without interfering with tsearch's
+		     * notion of the ordering of data.  That will create some
+		     * spurious empty data, so we add the changed() macro in a
+		     * few places to skip over those.
+		     */
+		    if (use_tsearch) {
+			char *trim = new_string(data->name);
+			trim[len1] = EOS;
+			data = add_tsearch_data(trim, data->base);
+			free(trim);
+			local = 1;
+		    } else
+#endif
+			data->name[len1] = EOS;
 		}
 	    } else if (len1 < len2) {
 		if (!strncmp(data->name, path, len1)) {
-		    TRACE(("trimming path '%s' to '%.*s'\n",
+		    TRACE(("** trimming path '%s' to '%.*s'\n",
 			   path, (int) len1, path));
 		    path[len2 = len1] = EOS;
 		}
@@ -566,21 +704,22 @@ do_merging(DATA * data, char *path, int *freed)
 		&& diff)
 		path += len2 - matched + 1;
 
-	    *freed = delink(data);
-	    TRACE(("merge @%d, prefix_opt=%d matched=%d diff=%d\n",
+	    if (!local)
+		*freed = delink(data);
+	    TRACE(("** merge @%d, prefix_opt=%d matched=%d diff=%d\n",
 		   __LINE__, prefix_opt, matched, diff));
 	} else if (!can_be_merged(path)) {
-	    TRACE(("do not merge, retain @%d\n", __LINE__));
+	    TRACE(("** do not merge, retain @%d\n", __LINE__));
 	    /* must not merge, retain existing name */
 	    path = data->name;
 	} else {
-	    TRACE(("merge @%d\n", __LINE__));
+	    TRACE(("** merge @%d\n", __LINE__));
 	    *freed = delink(data);
 	}
     } else if (!can_be_merged(path)) {
 	path = data->name;
     }
-    TRACE(("...do_merging %s\n", path));
+    TRACE(("** finish do_merging ->%s\n", path));
     return path;
 }
 
@@ -589,10 +728,10 @@ begin_data(const DATA * p)
 {
     if (!can_be_merged(p->name)
 	&& strchr(p->name, PATHSEP) != 0) {
-	TRACE(("begin_data:HAVE_PATH\n"));
+	TRACE(("** begin_data:HAVE_PATH\n"));
 	return HAVE_PATH;
     }
-    TRACE(("begin_data:HAVE_GENERIC\n"));
+    TRACE(("** begin_data:HAVE_GENERIC\n"));
     return HAVE_GENERIC;
 }
 
@@ -677,11 +816,97 @@ get_line(char **buffer, size_t *have, FILE *fp)
     return (used != 0);
 }
 
-#define date_delims(a,b) (((a)=='/' && (b)=='/') || ((a) == '-' && (b) == '-'))
+static char *
+data_filename(const DATA * p)
+{
+    return (p->name + (prefix_opt >= 0 ? p->base : prefix_len));
+}
+
+/*
+ * Count the (new)lines in a file, return -1 if the file is not found.
+ */
+static int
+count_lines(DATA * p)
+{
+    int result = -1;
+    char *filename = 0;
+    char *filetail = data_filename(p);
+    unsigned want = strlen(path_opt) + 2 + strlen(filetail);
+    FILE *fp;
+    int ch;
+
+    if ((filename = malloc(want)) != 0) {
+	sprintf(filename, "%s/%s", path_opt, filetail);
+	TRACE(("count_lines %s\n", filename));
+	if ((fp = fopen(filename, "r")) != 0) {
+	    result = 0;
+	    while ((ch = MY_FGETC(fp)) != EOF) {
+		if (ch == '\n')
+		    ++result;
+	    }
+	    fclose(fp);
+	} else {
+	    fprintf(stderr, "Cannot open %s\n", filename);
+	}
+	free(filename);
+    } else {
+	failed("count_lines");
+    }
+    return result;
+}
 
 static void
-do_file(FILE *fp)
+update_chunk(DATA * p, Change change)
 {
+    if (merge_opt) {
+	p->pending += 1;
+	p->chunk[change] += 1;
+    } else {
+	p->count[change] += 1;
+    }
+}
+
+static void
+finish_chunk(DATA * p)
+{
+    int i;
+
+    if (p->pending) {
+	p->pending = 0;
+	p->chunks += 1;
+	if (merge_opt) {
+	    /*
+	     * This is crude, but to make it really precise we would have
+	     * to keep an array of line-numbers to which which in a chunk
+	     * are marked as insert/delete.
+	     */
+	    if (p->chunk[cInsert] && p->chunk[cDelete]) {
+		int change;
+		if (p->chunk[cInsert] > p->chunk[cDelete]) {
+		    change = p->chunk[cDelete];
+		} else {
+		    change = p->chunk[cInsert];
+		}
+		p->chunk[cInsert] -= change;
+		p->chunk[cDelete] -= change;
+		p->chunk[cModify] += change;
+	    }
+	}
+	for_each_mark(i) {
+	    p->count[i] += p->chunk[i];
+	    p->chunk[i] = 0;
+	}
+    }
+}
+
+#define date_delims(a,b) (((a)=='/' && (b)=='/') || ((a) == '-' && (b) == '-'))
+#define CASE_TRACE() TRACE(("** handle case for '%c' %d:%s\n", *buffer, ok, that ? that->name : ""))
+
+static void
+do_file(FILE *fp, char *default_name)
+{
+    static const char *only_stars = "***************";
+
     DATA dummy;
     DATA *that = &dummy;
     DATA *prev = 0;
@@ -694,13 +919,20 @@ do_file(FILE *fp)
     size_t fixed = 0;
     int ok = HAVE_NOTHING;
     int marker = -1;
-    int unified = 0;
     int freed = 0;
+
+    int unified = 0;
     int old_unify = 0;
     int new_unify = 0;
+    int expect_unify = 0;
+
+    int old_dft = 0;
+    int new_dft = 0;
+
     int context = 1;
+
     char *s;
-#ifdef DEBUG
+#if OPT_TRACE
     int line_no = 0;
 #endif
 
@@ -726,15 +958,26 @@ do_file(FILE *fp)
 	}
 
 	/*
-	 * Trim trailing blanks (e.g., newline)
+	 * Trim trailing newline.
 	 */
 	for (s = buffer + strlen(buffer); s > buffer; s--) {
-	    if (isspace(UC(s[-1])))
+	    if ((UC(s[-1]) == '\n') || (UC(s[-1]) == '\r'))
 		s[-1] = EOS;
 	    else
 		break;
 	}
-	TRACE(("[%05d] %s\n", ++line_no, buffer));
+	++line_no;
+	TRACE(("[%05d] %s\n", line_no, buffer));
+
+	/*
+	 * "patch -U" can create ".rej" files lacking a filename header,
+	 * in unified format.  Check for those.
+	 */
+	if (line_no == 1 && !strncmp(buffer, "@@", 2)) {
+	    unified = 2;
+	    that = find_data(default_name);
+	    ok = begin_data(that);
+	}
 
 	/*
 	 * The lines identifying files in a context diff depend on how it was
@@ -750,9 +993,15 @@ do_file(FILE *fp)
 	 * lines to ensure we do not confuse the marker lines.
 	 */
 	marker = -1;
-	if (that != &dummy && !strcmp(buffer, "***************")) {
-	    TRACE(("begin context chunk\n"));
+	if (that != &dummy && !strcmp(buffer, only_stars)) {
+	    finish_chunk(that);
+	    TRACE(("** begin context chunk\n"));
 	    context = 2;
+	} else if (line_no == 1 && !strcmp(buffer, only_stars)) {
+	    TRACE(("** begin context chunk\n"));
+	    context = 2;
+	    that = find_data(default_name);
+	    ok = begin_data(that);
 	} else if (context == 2 && match(buffer, "*** ")) {
 	    context = 1;
 	} else if (context == 1 && match(buffer, "--- ")) {
@@ -761,11 +1010,13 @@ do_file(FILE *fp)
 	} else if (match(buffer, "*** ")) {
 	    marker = 0;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "--- ")) {
+	    finish_chunk(that);
 	    marker = unified = 1;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "+++ ")) {
 	    marker = unified = 2;
 	} else if (unified == 2
 		   || ((old_unify + new_unify) == 0 && (*buffer == '@'))) {
+	    finish_chunk(that);
 	    unified = 0;
 	    if (*buffer == '@') {
 		int old_base, new_base, old_size, new_size;
@@ -795,17 +1046,19 @@ do_file(FILE *fp)
 	     * "--" in the first two columns of the diff'd file.
 	     */
 	    unified = 0;
-	    TRACE(("Expected \"+++\" @%d:%s\n", __LINE__, buffer));
+	    TRACE(("?? Expected \"+++\" for unified diff\n"));
 	    if (prev != 0
 		&& prev != that
 		&& InsOf(that) == 0
 		&& DelOf(that) == 0
 		&& strcmp(prev->name, that->name)) {
-		TRACE(("giveup on %ld/%ld %s\n", InsOf(that), DelOf(that), that->name));
-		TRACE(("revert to %ld/%ld %s\n", InsOf(prev), DelOf(prev), prev->name));
+		TRACE(("?? giveup on %ld/%ld %s\n", InsOf(that),
+		       DelOf(that), that->name));
+		TRACE(("?? revert to %ld/%ld %s\n", InsOf(prev),
+		       DelOf(prev), prev->name));
 		(void) delink(that);
 		that = prev;
-		DelOf(that) += 1;
+		update_chunk(that, cDelete);
 	    }
 	} else if (old_unify + new_unify) {
 	    switch (*buffer) {
@@ -817,6 +1070,7 @@ do_file(FILE *fp)
 		if (new_unify)
 		    --new_unify;
 		break;
+	    case '\0':
 	    case ' ':
 		if (old_unify)
 		    --old_unify;
@@ -824,11 +1078,44 @@ do_file(FILE *fp)
 		    --new_unify;
 		break;
 	    default:
+		TRACE(("?? expected more in chunk\n"));
 		old_unify = new_unify = 0;
 		break;
 	    }
+	    if (!(old_unify + new_unify)) {
+		expect_unify = 2;
+	    }
 	} else {
+	    int old_base, new_base;
+
 	    unified = 0;
+
+	    if (line_no == 1
+		&& decode_default(buffer,
+				  &old_base, &old_dft,
+				  &new_base, &new_dft)) {
+		TRACE(("DFT %d,%d -> %d,%d\n",
+		       old_base, old_base + old_dft - 1,
+		       new_base, new_base + new_dft - 1));
+		finish_chunk(that);
+		that = find_data("unknown");
+		ok = begin_data(that);
+	    }
+	}
+
+	/*
+	 * If the previous line ended a chunk of a unified diff, we may begin
+	 * another chunk, or begin another type of diff.  If neither, do not
+	 * continue to accumulate counts for the unified diff which has ended.
+	 */
+	if (expect_unify != 0) {
+	    if (expect_unify-- == 1) {
+		if (unified == 0) {
+		    TRACE(("?? did not get chunk\n"));
+		    finish_chunk(that);
+		    that = &dummy;
+		}
+	    }
 	}
 
 	/*
@@ -836,7 +1123,7 @@ do_file(FILE *fp)
 	 * below.
 	 */
 	if (marker > 0) {
-	    TRACE(("@%d, marker=%d, override %s\n", __LINE__, marker, buffer));
+	    TRACE(("** have marker=%d, override %s\n", marker, buffer));
 	    (void) strncpy(buffer, "***", 3);
 	}
 
@@ -846,6 +1133,7 @@ do_file(FILE *fp)
 	 */
 	switch (*buffer) {
 	case 'O':		/* Only */
+	    CASE_TRACE();
 	    if (match(buffer, "Only in ")) {
 		char *path = buffer + 8;
 		int found = 0;
@@ -860,6 +1148,7 @@ do_file(FILE *fp)
 		}
 		if (found) {
 		    blip('.');
+		    finish_chunk(that);
 		    that = find_data(path);
 		    that->cmt = Only;
 		    ok = HAVE_NOTHING;
@@ -873,11 +1162,13 @@ do_file(FILE *fp)
 	     * pathname of the files; some put only the leaf names.
 	     */
 	case 'I':
+	    CASE_TRACE();
 	    if (match(buffer, "Index: ")) {
 		s = strrchr(buffer, BLANK);	/* last token is name */
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
+		finish_chunk(that);
 		s = do_merging(that, s, &freed);
 		that = find_data(s);
 		ok = begin_data(that);
@@ -885,12 +1176,14 @@ do_file(FILE *fp)
 	    break;
 
 	case 'd':		/* diff command trace */
+	    CASE_TRACE();
 	    if (match(buffer, "diff ")
 		&& *(s = skip_options(buffer + 5)) != '\0') {
 		s = strrchr(buffer, BLANK);
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
+		finish_chunk(that);
 		s = do_merging(that, s, &freed);
 		that = find_data(s);
 		ok = begin_data(that);
@@ -898,7 +1191,7 @@ do_file(FILE *fp)
 	    break;
 
 	case '*':
-	    TRACE(("@%d, ok=%d:%s\n", __LINE__, ok, buffer));
+	    CASE_TRACE();
 	    if (!(ok & HAVE_PATH)) {
 		int ddd, hour, minute, second;
 		int day, month, year;
@@ -942,12 +1235,13 @@ do_file(FILE *fp)
 			&& !edit_range(b_fname))
 		    ) {
 		    prev = that;
+		    finish_chunk(that);
 		    s = do_merging(that, b_fname, &freed);
 		    if (freed)
 			prev = 0;
 		    that = find_data(s);
 		    ok = begin_data(that);
-		    TRACE(("after merge:%d:%s\n", ok, s));
+		    TRACE(("** after merge:%d:%s\n", ok, s));
 		}
 	    }
 	    break;
@@ -955,36 +1249,48 @@ do_file(FILE *fp)
 	case '+':
 	    /* FALL-THRU */
 	case '>':
-	    if (ok)
-		InsOf(that) += 1;
+	    CASE_TRACE();
+	    if (ok) {
+		update_chunk(that, cInsert);
+	    }
 	    break;
 
 	case '-':
-	    if (!ok)
+	    if (!ok) {
+		CASE_TRACE();
 		break;
-	    if (!unified && !strcmp(buffer, "---"))
+	    }
+	    if (!unified && !strcmp(buffer, "---")) {
+		CASE_TRACE();
 		break;
+	    }
 	    /* fall-thru */
 	case '<':
-	    if (ok)
-		DelOf(that) += 1;
+	    CASE_TRACE();
+	    if (ok) {
+		update_chunk(that, cDelete);
+	    }
 	    break;
 
 	case '!':
-	    if (ok)
-		ModOf(that) += 1;
+	    CASE_TRACE();
+	    if (ok) {
+		update_chunk(that, cModify);
+	    }
 	    break;
 
 	    /* Expecting "Binary files XXX and YYY differ" */
 	case 'B':		/* Binary */
 	    /* FALL-THRU */
 	case 'b':		/* binary */
+	    CASE_TRACE();
 	    if (match(buffer + 1, "inary files ")) {
 		s = strrchr(buffer, BLANK);
 		if (!strcmp(s, " differ")) {
 		    *s = EOS;
 		    s = strrchr(buffer, BLANK);
 		    blip('.');
+		    finish_chunk(that);
 		    that = find_data(skip_blanks(s));
 		    that->cmt = Binary;
 		    ok = HAVE_NOTHING;
@@ -995,6 +1301,8 @@ do_file(FILE *fp)
     }
     blip('\n');
 
+    finish_chunk(that);
+    finish_chunk(&dummy);
     if (buffer != 0) {
 	free(buffer);
 	free(b_fname);
@@ -1049,7 +1357,7 @@ plot_round1(const long num[MARKS])
     long half = (plot_scale / 2);
     int i, j;
 
-    for (i = 0; i < MARKS; ++i) {
+    for_each_mark(i) {
 	long product = (plot_width * num[i]);
 	scaled[i] = (product / plot_scale);
 	remain[i] = (product % plot_scale);
@@ -1057,7 +1365,8 @@ plot_round1(const long num[MARKS])
 	have += product - remain[i];
     }
     while (want > have) {
-	for (i = 0, j = -1; i < MARKS; ++i) {
+	j = -1;
+	for_each_mark(i) {
 	    if (remain[i] != 0
 		&& (remain[i] > (j >= 0 ? remain[j] : half))) {
 		j = i;
@@ -1071,7 +1380,7 @@ plot_round1(const long num[MARKS])
 	    break;
 	}
     }
-    for (i = 0; i < MARKS; ++i) {
+    for_each_mark(i) {
 	plot_bar(scaled[i], marks[i]);
 	result += scaled[i];
     }
@@ -1089,10 +1398,10 @@ plot_round2(const long num[MARKS])
     long result = 0;
     long scaled[MARKS];
     long remain[MARKS];
-    long total;
+    long total = 0;
     int i;
 
-    for (total = 0, i = 0; i < MARKS; i++)
+    for (i = 0; i < MARKS; i++)
 	total += num[i];
 
     if (total == 0)
@@ -1103,7 +1412,7 @@ plot_round2(const long num[MARKS])
     if (total == 0)
 	total++;
 
-    for (i = 0; i < MARKS; i++) {
+    for_each_mark(i) {
 	scaled[i] = num[i] * plot_width / plot_scale;
 	remain[i] = num[i] * plot_width - scaled[i] * plot_scale;
 	total -= scaled[i];
@@ -1117,7 +1426,7 @@ plot_round2(const long num[MARKS])
 	/* search for the largest remainder */
 	largest = largest_count = 0;
 	max_remain = 0;
-	for (i = 0; i < MARKS; i++) {
+	for_each_mark(i) {
 	    if (remain[i] > max_remain) {
 		largest = 1 << i;
 		largest_count = 1;
@@ -1134,7 +1443,7 @@ plot_round2(const long num[MARKS])
 	    break;
 
 	/* allocate the extra characters */
-	for (i = 0; i < MARKS; i++) {
+	for_each_mark(i) {
 	    if (largest & (1 << i)) {
 		scaled[i]++;
 		total--;
@@ -1143,8 +1452,9 @@ plot_round2(const long num[MARKS])
 	}
     }
 
-    for (i = 0; i < MARKS; i++)
+    for_each_mark(i) {
 	result += plot_bar(scaled[i], marks[i]);
+    }
 
     return result;
 }
@@ -1162,17 +1472,20 @@ plot_numbers(const DATA * p)
 	printf("%5ld ", InsOf(p));
 	printf("%5ld ", DelOf(p));
 	printf("%5ld ", ModOf(p));
+	if (path_opt)
+	    printf("%5ld ", EqlOf(p));
     }
 
     if (format_opt == FMT_CONCISE) {
-	for (i = 0; i < MARKS; i++) {
+	for_each_mark(i) {
 	    printf("\t%ld %c", p->count[i], marks[i]);
 	}
     } else {
 	switch (round_opt) {
 	default:
-	    for (i = 0; i < MARKS; ++i)
+	    for_each_mark(i) {
 		used += plot_num(p->count[i], marks[i], &temp);
+	    }
 	    break;
 	case 1:
 	    used = plot_round1(p->count);
@@ -1192,20 +1505,30 @@ plot_numbers(const DATA * p)
     }
 }
 
+#define changed(p) (!merge_names \
+		    || (p)->cmt != Normal \
+		    || (TotalOf(p)) != 0)
+
 static void
 show_data(const DATA * p)
 {
-    char *name = p->name + (prefix_opt >= 0 ? p->base : prefix_len);
+    char *name = data_filename(p);
 
-    if (table_opt) {
+    if (!changed(p)) {
+	;
+    } else if (p->cmt == Binary && suppress_binary == 1) {
+	;
+    } else if (table_opt) {
 	if (names_only) {
 	    printf("%s\n", name);
 	} else {
-	    printf("%ld,%ld,%ld,%s\n",
+	    printf("%ld,%ld,%ld,",
 		   InsOf(p),
 		   DelOf(p),
-		   ModOf(p),
-		   name);
+		   ModOf(p));
+	    if (path_opt)
+		printf("%ld,", EqlOf(p));
+	    printf("%s\n", name);
 	}
     } else if (names_only) {
 	printf("%s\n", name);
@@ -1241,18 +1564,34 @@ show_tsearch(const void *nodep, const VISIT which, const int depth)
 }
 #endif
 
+static int
+ignore_data(DATA * p)
+{
+    return ((!changed(p))
+	    || (p->cmt == Binary && suppress_binary));
+}
+
 static void
 summarize(void)
 {
     DATA *p;
-    long total_ins = 0, total_del = 0, total_mod = 0, temp;
+    long total_ins = 0;
+    long total_del = 0;
+    long total_mod = 0;
+    long total_eql = 0;
+    long temp;
     int num_files = 0, shortest_name = -1, longest_name = -1;
 
     plot_scale = 0;
     for (p = all_data; p; p = p->link) {
 	int len = strlen(p->name);
 
+	if (ignore_data(p))
+	    continue;
+
 	/*
+	 * If "-pX" option is given, prefix_opt is positive.
+	 *
 	 * "-p0" gives the whole pathname unmodified.  "-p1" strips
 	 * through the first path-separator, etc.
 	 */
@@ -1261,6 +1600,10 @@ summarize(void)
 	    if (name_wide < (len - p->base))
 		name_wide = (len - p->base);
 	} else {
+	    /*
+	     * If "-pX" option is not given, strip off any prefix which is
+	     * shared by all of the names.
+	     */
 	    if (len < prefix_len || prefix_len < 0)
 		prefix_len = len;
 	    while (prefix_len > 0) {
@@ -1277,14 +1620,41 @@ summarize(void)
 	    if (len < shortest_name || shortest_name < 0)
 		shortest_name = len;
 	}
+    }
 
-	num_files++;
-	total_ins += InsOf(p);
-	total_del += DelOf(p);
-	total_mod += ModOf(p);
-	temp = TotalOf(p);
-	if (temp > plot_scale)
-	    plot_scale = temp;
+    /*
+     * Use a separate loop after computing prefix_len so we can apply the "-S"
+     * or "-D" options to find files that we can use as reference for the
+     * unchanged-count.
+     */
+    for (p = all_data; p; p = p->link) {
+	if (!ignore_data(p)) {
+	    EqlOf(p) = 0;
+	    if (path_opt != 0) {
+		int count = count_lines(p);
+
+		if (count >= 0) {
+		    EqlOf(p) = count - ModOf(p);
+		    if (path_dest) {
+			EqlOf(p) -= InsOf(p);
+			InsOf(p) = 0;
+		    } else {
+			EqlOf(p) -= DelOf(p);
+			DelOf(p) = 0;
+		    }
+		    if (EqlOf(p) < 0)
+			EqlOf(p) = 0;
+		}
+	    }
+	    num_files++;
+	    total_ins += InsOf(p);
+	    total_del += DelOf(p);
+	    total_mod += ModOf(p);
+	    total_eql += EqlOf(p);
+	    temp = TotalOf(p);
+	    if (temp > plot_scale)
+		plot_scale = temp;
+	}
     }
 
     if (prefix_opt < 0) {
@@ -1302,10 +1672,14 @@ summarize(void)
     if (plot_scale < plot_width)
 	plot_scale = plot_width;	/* 1:1 */
 
-    if (table_opt)
-	printf("%sFILENAME\n",
-	       (names_only ? "" : "INSERTED,DELETED,MODIFIED,"));
-
+    if (table_opt) {
+	if (!names_only) {
+	    printf("INSERTED,DELETED,MODIFIED,");
+	    if (path_opt)
+		printf("UNCHANGED,");
+	}
+	printf("FILENAME\n");
+    }
 #ifdef HAVE_TSEARCH
     if (use_tsearch) {
 	twalk(sorted_data, show_tsearch);
@@ -1317,14 +1691,18 @@ summarize(void)
 
     if (!table_opt && !names_only) {
 #define PLURAL(n) n, n != 1 ? "s" : ""
-	printf("%s %d file%s changed", comment_opt, PLURAL(num_files));
-	if (total_ins)
-	    printf(", %ld insertion%s(+)", PLURAL(total_ins));
-	if (total_del)
-	    printf(", %ld deletion%s(-)", PLURAL(total_del));
-	if (total_mod)
-	    printf(", %ld modification%s(!)", PLURAL(total_mod));
-	(void) putchar('\n');
+	if (num_files > 0 || !quiet) {
+	    printf("%s %d file%s changed", comment_opt, PLURAL(num_files));
+	    if (total_ins)
+		printf(", %ld insertion%s(+)", PLURAL(total_ins));
+	    if (total_del)
+		printf(", %ld deletion%s(-)", PLURAL(total_del));
+	    if (total_mod)
+		printf(", %ld modification%s(!)", PLURAL(total_mod));
+	    if (total_eql && path_opt != 0)
+		printf(", %ld unchanged line%s(=)", PLURAL(total_eql));
+	    (void) putchar('\n');
+	}
     }
 }
 
@@ -1345,13 +1723,28 @@ is_compressed(const char *name)
     }
     if (verb != 0 && *verb != '\0') {
 	result = (char *) xmalloc(strlen(verb) + 10 + len);
-	sprintf(result, "%s -dc %s", verb, name);
+	sprintf(result, "%s -dc \"%s\"", verb, name);
     }
     return result;
 
 }
 #endif
 
+static void
+set_path_opt(char *value, int destination)
+{
+    path_opt = value;
+    path_dest = destination;
+    if (*path_opt != 0) {
+	if (is_dir(path_opt)) {
+	    num_marks = 4;
+	} else {
+	    fprintf(stderr, "Not a directory:%s\n", path_opt);
+	    exit(EXIT_FAILURE);
+	}
+    }
+}
+
 static void
 usage(FILE *fp)
 {
@@ -1365,18 +1758,25 @@ usage(FILE *fp)
 	"",
 	"Options:",
 	"  -c      prefix each line with comment (#)",
+#if OPT_TRACE
+	"  -d      debug - prints a lot of information",
+#endif
+	"  -D PATH specify location of patched files, use for unchanged-count",
 	"  -e FILE redirect standard error to FILE",
 	"  -f NUM  format (0=concise, 1=normal, 2=filled, 4=values)",
 	"  -h      print this message",
 	"  -k      do not merge filenames",
 	"  -l      list filenames only",
+	"  -m      merge insert/delete data in chunks as modified-lines",
 	"  -n NUM  specify minimum width for the filenames (default: auto)",
 	"  -o FILE redirect standard output to FILE",
 	"  -p NUM  specify number of pathname-separators to strip (default: common)",
+	"  -q      suppress the \"0 files changed\" message for empty diffs",
 	"  -r NUM  specify rounding for histogram (0=none, 1=simple, 2=adjusted)",
+	"  -S PATH specify location of original files, use for unchanged-count",
 	"  -t      print a table (comma-separated-values) rather than histogram",
 	"  -u      do not sort the input list",
-	"  -v      makes output more verbose",
+	"  -v      show progress if output is redirected to a file",
 	"  -V      prints the version number",
 	"  -w NUM  specify maximum width of the output (default: 80)",
     };
@@ -1413,12 +1813,21 @@ main(int argc, char *argv[])
 
     max_width = 80;
 
-    while ((j = getopt_helper(argc, argv, "ce:f:hkln:o:p:r:tuvVw:", 'h', 'V'))
+    while ((j = getopt_helper(argc, argv,
+			      "bcdD:e:f:hklmn:o:p:qr:S:tuvVw:", 'h', 'V'))
 	   != -1) {
 	switch (j) {
+	case 'b':
+	    suppress_binary = 1;
+	    break;
 	case 'c':
 	    comment_opt = "#";
 	    break;
+#if OPT_TRACE
+	case 'd':
+	    trace_opt = 1;
+	    break;
+#endif
 	case 'e':
 	    if (freopen(optarg, "w", stderr) == 0)
 		failed(optarg);
@@ -1435,6 +1844,9 @@ main(int argc, char *argv[])
 	case 'l':
 	    names_only = 1;
 	    break;
+	case 'm':
+	    merge_opt = 1;
+	    break;
 	case 'n':
 	    name_wide = atoi(optarg);
 	    break;
@@ -1445,6 +1857,12 @@ main(int argc, char *argv[])
 	case 'p':
 	    prefix_opt = atoi(optarg);
 	    break;
+	case 'D':
+	    set_path_opt(optarg, 1);
+	    break;
+	case 'S':
+	    set_path_opt(optarg, 0);
+	    break;
 	case 'r':
 	    round_opt = atoi(optarg);
 	    break;
@@ -1467,11 +1885,22 @@ main(int argc, char *argv[])
 	case 'w':
 	    max_width = atoi(optarg);
 	    break;
+	case 'q':
+	    quiet = 1;
+	    break;
 	default:
 	    usage(stderr);
 	    return (EXIT_FAILURE);
 	}
     }
+
+    /*
+     * The numbers from -S/-D options will only be useful if the merge option
+     * is added.
+     */
+    if (path_opt)
+	merge_opt = 1;
+
     show_progress = verbose && (!isatty(fileno(stdout))
 				&& isatty(fileno(stderr)));
 
@@ -1491,7 +1920,7 @@ main(int argc, char *argv[])
 			(void) fprintf(stderr, "%s\n", name);
 			(void) fflush(stderr);
 		    }
-		    do_file(fp);
+		    do_file(fp, name);
 		    (void) pclose(fp);
 		}
 		free(command);
@@ -1502,17 +1931,17 @@ main(int argc, char *argv[])
 		    (void) fprintf(stderr, "%s\n", name);
 		    (void) fflush(stderr);
 		}
-		do_file(fp);
+		do_file(fp, name);
 		(void) fclose(fp);
 	    } else {
 		failed(name);
 	    }
 	}
     } else {
-	do_file(stdin);
+	do_file(stdin, "unknown");
     }
     summarize();
-#if defined(DEBUG) || defined(NO_LEAKS)
+#if defined(NO_LEAKS)
     while (all_data != 0) {
 	delink(all_data);
     }
